#ifndef __COMMCONNECT_H
#define __COMMCONNECT_H

#include <dlfcn.h>
#include <string>
#include <vector>
#include <list>
#include <set>
#include <iostream>
#include <pthread.h>

#include <sys/ioctl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>

#include "adapter.h"
#include "property.h"

/*
#define COMMDEBUG
 */

#define NO_THREAD_ADAPTER
#define IGNORE_READ_WRITE_PROTECTED_

#if defined(__QNX__)
#define CRTSCTS (IHFLOW | OHFLOW)
#endif

using namespace std;

class TPortException
{
    protected:
        std::string _Message;
    public:
        TPortException(std::string Message);
        TPortException(std::string Location, std::string Message);
        TPortException(const TPortException &e);
        std::string Message(void) const;
        friend std::ostream& operator<< (std::ostream &os, const TPortException &e);
};

class TConnection;
typedef void (*TConnectionNotifyEvent)(TConnection*);

class TConnection
{
    private:
        bool FActive;
        TConnectionNotifyEvent FBeforeOpen, FBeforeClose, FAfterOpen, FAfterClose;

        bool get_Active();
        void set_Active(bool aEnable);
        TConnectionNotifyEvent get_BeforeOpen();
        void set_BeforeOpen(TConnectionNotifyEvent value);
        TConnectionNotifyEvent get_AfterOpen();
        void set_AfterOpen(TConnectionNotifyEvent value);
        TConnectionNotifyEvent get_BeforeClose();
        void set_BeforeClose(TConnectionNotifyEvent value);
        TConnectionNotifyEvent get_AfterClose();
        void set_AfterClose(TConnectionNotifyEvent value);

    protected:

        virtual bool OpenConn() {return true;};
        virtual void CloseConn()=0;
        virtual void DoBeforeOpen();
        virtual void DoAfterOpen();
        virtual void DoBeforeClose();
        virtual void DoAfterClose();
        void CheckActive();
        void CheckInactive();

    public:
        TConnection();
        virtual ~TConnection();
        void Open();
        void Close();
        property <bool, TConnection> Active;
        property <TConnectionNotifyEvent, TConnection> BeforeOpen;
        property <TConnectionNotifyEvent, TConnection> AfterOpen;
        property <TConnectionNotifyEvent, TConnection> BeforeClose;
        property <TConnectionNotifyEvent, TConnection> AfterClose;

};

typedef void (*TOnExceptionEvent)(TConnection* Sender, TPortException& E);

class TLogger : public TConnection
{
    private:
        bool FAutoOpen;
        TOnExceptionEvent FOnException;
        bool get_AutoOpen();
        void set_AutoOpen(bool aValue);
        TOnExceptionEvent get_OnException();
        void set_OnException(TOnExceptionEvent value);

    protected:
        pthread_mutex_t* FCriticalSection;
        virtual void DoOnException (TPortException& E);
    public:
        TLogger();
// virtual as well as TConnection's destructor.
        ~TLogger();
        property <bool, TLogger> AutoOpen;
        property <TOnExceptionEvent, TLogger> OnException;

};

class TLogConnection : public TConnection
{
    private:
        TLogger* FLogger;
    public:
        TLogConnection() : TConnection () { };
};

class TCommunicationConnection;
typedef void (*TCommRxCharEvent)(TCommunicationConnection* Sender, int Count, void* DataBind);
class TCommunicationConnection : public TLogConnection
{
    private:
        TCommRxCharEvent FOnRxChar;
        TCommRxCharEvent get_OnRxChar();
        void set_OnRxChar(TCommRxCharEvent value);
    protected:
        void* FDataBind;
        virtual int Write (unsigned char* Buf, int Count) = 0;
        virtual int Read (unsigned char* Buf, int Count) = 0;
        virtual void DoOnRxChar(int Count);

    public:
        TCommunicationConnection();
        int Send(std::string S);
        virtual int InQueCount() = 0;
        std::string Retrieve(int aCount);
        virtual void PurgeIn() = 0;
        virtual void PurgeOut() = 0;
        virtual void Bind (void* DataBind) = 0;
        property <TCommRxCharEvent, TCommunicationConnection> OnRxChar;
};

typedef int TComHandle;

const unsigned long INFINITE = 0xFFFFFFFF;
const int INVALID_HANDLE_VALUE = -1;

typedef void (*TCommEvent)(TConnection* Sender, unsigned long Status);
enum TCommEventType
{
    evBreak = 0,
    evCts,
    evDsr,
    evError,
    evRing,
    evRlsd,
    evRxChar,
    evRxFlag,
    evTxEmpty,
    evTxChar
};

enum TBaudrate
{
    br110 = 0,
    br300,
    br600,
    br1200,
    br2400,
    br4800,
    br9600,
    br14400,
    br19200,
    br38400,
    br56000,
    br57600,
    br115200,
    br128000,
    br256000
};

enum TParity
{
    paNone = 0,
    paOdd,
    paEven,
    paMark,
    paSpace
};

enum TStopbits
{
    sb10 = 0,
    sb15,
    sb20
};

enum TDatabits
{
    da4 = 0,
    da5,
    da6,
    da7,
    da8 = 4
};

enum TFlowControl
{
    fcNone = 0,
    fcCTS,
    fcDTR,
    fcSoftware,
    fcDefault
};

enum TCommOption
{
    coParityCheck = 0,
    coDsrSensitivity,
    coIgnoreXOff,
    coErrorChar,
    coNullStrip
};

typedef set<enum TCommOption> TCommOptions;
typedef set<enum TCommEventType> TCommEventTypes;

typedef void (*TNotifyEvent)(TConnection* Sender);

class TCommHandle;

typedef void (*TCommErrorEvent)(TConnection* Sender, int Errors);

typedef class TCommEventThread
{
    private:
        TComHandle get_CommHandle();
        bool get_Terminated();
        pthread_mutex_t* get_CriticalSection();
        TCommHandle* get_Comm();
    protected:
        volatile bool FTerminated;
        TComHandle FCommHandle;
        pthread_mutex_t* FCriticalSection;

        TCommEventTypes FEvents;
        int FModemFlags;

        TCommHandle* FComm;

    public:
        rproperty <TComHandle, TCommEventThread> CommHandle;
        rproperty <bool, TCommEventThread> Terminated;
        rproperty <pthread_mutex_t*, TCommEventThread> CriticalSection;
        rproperty <TCommHandle*, TCommEventThread> Comm;

#ifdef __CYGWIN__
        volatile bool FConsiderTerminated;
        pthread_mutex_t FConsiderTerminatedMutex;
#endif

        unsigned long EventMask;
        void DoOnSignal();

        TCommEventThread(TCommHandle* aComm, TComHandle Handle, TCommEventTypes Events);
        ~TCommEventThread();
        bool WriteFlag;
        void Terminate();
#ifdef __CYGWIN__
        bool ConsiderTerminated();
#endif

        void* run();
};

class TCommHandle : public TCommunicationConnection
{
    private:
        TBaudrate FBaudrate;
        TParity FParity;
        TStopbits FStopbits;
        TDatabits FDatabits;
        TFlowControl FFlowControl;
        TCommOptions FOptions;
        int FReadTimeout;
        int FWriteTimeout;
        int FReadBufSize;
        int FWriteBufSize;
        TCommEventTypes FMonitorEvents;
        char FXonChar;
        char FXoffChar;
        char FErrorChar;
        char FEofChar;
        char FEvtChar;

        pthread_mutex_t* FCriticalSection;
        TCommEventThread* FEventThread;
        TNotifyEvent FOnBreak;
        TNotifyEvent FOnCts;
        TNotifyEvent FOnDsr;
        TCommErrorEvent FOnError;
        TNotifyEvent FOnRing;
        TNotifyEvent FOnRlsd;
        TNotifyEvent FOnRxFlag;
        TNotifyEvent FOnTxEmpty;
        TNotifyEvent FOnTxChar;

        TComHandle get_hCommDev();
        void SethCommDev(TComHandle value);

        TBaudrate get_Baudrate();
        void set_Baudrate(TBaudrate value);
        TParity get_Parity();
        void set_Parity(TParity value);
        TStopbits get_Stopbits();
        void set_Stopbits(TStopbits value);
        TDatabits get_Databits();
        void set_Databits(TDatabits value);
        TCommOptions get_Options();
        void set_Options(TCommOptions value);
        TFlowControl get_FlowControl();
        void set_FlowControl(TFlowControl value);

        char get_XonChar ();
        void set_XonChar (char value);
        char get_XoffChar ();
        void set_XoffChar (char value);
        char get_ErrorChar ();
        void set_ErrorChar (char value);
        char get_EofChar ();
        void set_EofChar (char value);
        char get_EvtChar ();
        void set_EvtChar (char value);

        TNotifyEvent get_OnBreak();
        void set_OnBreak(TNotifyEvent value);
        TNotifyEvent get_OnCts();
        void set_OnCts(TNotifyEvent value);
        TNotifyEvent get_OnDsr();
        void set_OnDsr(TNotifyEvent value);
        TNotifyEvent get_OnRing();
        void set_OnRing(TNotifyEvent value);
        TNotifyEvent get_OnRlsd();
        void set_OnRlsd(TNotifyEvent value);

        TCommErrorEvent get_OnError();
        void set_OnError(TCommErrorEvent value);

        TNotifyEvent get_OnRxFlag();
        void set_OnRxFlag(TNotifyEvent value);
        TNotifyEvent get_OnTxEmpty();
        void set_OnTxEmpty(TNotifyEvent value);
        TNotifyEvent get_OnTxChar();
        void set_OnTxChar(TNotifyEvent value);

        int get_ReadTimeout();
        void set_ReadTimeout(int value);
        int get_WriteTimeout();
        void set_WriteTimeout(int value);

        int get_ReadBufSize();
        void set_ReadBufSize(int value);
        int get_WriteBufSize();
        void set_WriteBufSize(int value);
        TCommEventTypes get_MonitorEvents();
        void set_MonitorEvents(TCommEventTypes value);

    protected:
        TComHandle FhCommDev;
        pthread_t t;
        void UpdateDataControlBlock();
        virtual void UpdateDCB();
        bool GetModemState (int index);
        void SetEscBreak(bool value);
        void SetEsc(int Index, bool value);
        void UpdateCommTimeouts();
        virtual bool OpenConn();
        virtual void CloseConn();
        unsigned long GetTickCount(void);
        bool IsErrorMonitored ();
#ifndef IGNORE_READ_WRITE_PROTECTED_
        virtual int Write (unsigned char* Buf, int Count);
        virtual int Read (unsigned char* Buf, int Count);
#endif
        friend void HandleCommEvent(unsigned long status, TCommHandle* commhandle);

    public:
        TCommHandle();
        ~TCommHandle();
        rproperty <TComHandle, TCommHandle> hCommDev;
        property <TBaudrate, TCommHandle> Baudrate;
        property <TParity, TCommHandle> Parity;
        property <TStopbits, TCommHandle> Stopbits;
        property <TDatabits, TCommHandle> Databits;
        property <TCommOptions, TCommHandle> Options;
        property <TFlowControl, TCommHandle> FlowControl;
        property <char, TCommHandle> XonChar;
        property <char, TCommHandle> XoffChar;
        property <char, TCommHandle> ErrorChar;
        property <char, TCommHandle> EofChar;
        property <char, TCommHandle> EvtChar;
        property <int, TCommHandle> ReadTimeout;
        property <int, TCommHandle> WriteTimeout;
        property <int, TCommHandle> ReadBufSize;
        property <int, TCommHandle> WriteBufSize;
        property <TCommEventTypes, TCommHandle> MonitorEvents;
        property <TNotifyEvent, TCommHandle> OnBreak;
        property <TNotifyEvent, TCommHandle> OnCts;
        property <TNotifyEvent, TCommHandle> OnDsr;
        property <TNotifyEvent, TCommHandle> OnRing;
        property <TNotifyEvent, TCommHandle> OnRlsd;
        property <TCommErrorEvent, TCommHandle> OnError;
        property <TNotifyEvent, TCommHandle> OnRxFlag;
        property <TNotifyEvent, TCommHandle> OnTxEmpty;
        property <TNotifyEvent, TCommHandle> OnTxChar;
        virtual int InQueCount();
        int OutQueCount();
        void Lock();
        void Unlock();
        virtual void PurgeIn();
        virtual void PurgeOut();
        virtual void Bind (void* DataBind);

#ifdef IGNORE_READ_WRITE_PROTECTED_
        virtual int Write (unsigned char* Buf, int Count);
        virtual int Read (unsigned char* Buf, int Count);
#endif
        void SetDTRState(bool value);
        void SetRTSState(bool value);
};

#if defined(__QNX__)
const std::string DefaultDeviceName = "/dev/ser1";
#else
const std::string DefaultDeviceName = "/dev/ttyS0";
#endif

class TComm : public TCommHandle
{
    private:
        std::string FDeviceName;
        std::string get_DeviceName();
        void set_DeviceName(std::string value);
    protected:
        virtual bool OpenConn();
        virtual void CloseConn();
    public:
        TComm();
        ~TComm();
        property <std::string, TComm> DeviceName;
};

// COMMCONNECT_H
#endif
